﻿using BinDataLoader.Enumeration;
using BinDataLoader.Multiboot;
using Ecng.Xaml;
using Microsoft.Win32;
using StockSharp.Algo.Candles;
using StockSharp.Algo.History.Russian.Finam;
using StockSharp.Algo.Storages;
using StockSharp.BusinessEntities;
using StockSharp.Messages;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Threading;
using System.Xml.Serialization;

namespace BinDataLoader
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private StorageRegistry _storageRegistry;
        /// <summary>
        /// Сырые аргументы
        /// </summary>
        string[] rawArgs;

        Dictionary<KeyArg, string> args;
        TimeSpan candleTimeFrame;

        public MainWindow()
        {
            InitializeComponent();
            this.Title = "Bin Data Loader " + Assembly.GetExecutingAssembly().GetName().Version.ToString();

            candleTimeFrame = TimeSpan.FromMinutes(1);

            rawArgs = Environment.GetCommandLineArgs();
            args = new Dictionary<KeyArg, string>(); 
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            HistorySourceComboBox.Items.Add((HistorySource.Finam));
            HistorySourceComboBox.Items.Add((HistorySource.FinViz));
            HistorySourceComboBox.Items.Add((HistorySource.GoogleFinance));
            HistorySourceComboBox.Items.Add((HistorySource.Quandl));
            HistorySourceComboBox.Items.Add((HistorySource.Xignite));
            HistorySourceComboBox.Items.Add((HistorySource.YahooFinance));

            HistorySourceComboBox.SelectedIndex = 0;

            SecuritySourceComboBox.Items.Add("stocks.xml");
            SecuritySourceComboBox.Items.Add("futures.xml");

            SecuritySourceComboBox.SelectedIndex = 0;

            this.GuiAsync(() =>
            {
                BinStoragePathLabel.Content = @"F:\HistoricalData\";
            });

            //Хранилище, через которое будет производиться доступ к тиковой и котировочной базе
            _storageRegistry = new StorageRegistry
            {
                // set historical path
                DefaultDrive = new LocalMarketDataDrive((string)BinStoragePathLabel.Content)
            };

            if (rawArgs.Length == 1) // запуск app без параметров
            {
                StartDatePicker.SelectedDate = new DateTime(2017, 4, 1);
                EndDatePicker.SelectedDate = new DateTime(2017, 4, 7);

                UseMultibootMenuItem.IsChecked = false;
            }
            else // запуск app с параметрами
            {
                Dictionary<TimeFrame, int> date;
                TestType testType = TestType.Candle;

                List<KeyArg> keysArg = new List<KeyArg>()
                {
                    // последовательность ключей
                    KeyArg.Security, // -security
                    KeyArg.From,     // -from (gggg, m, d)
                    KeyArg.To,       // -to (gggg, m, d)
                    KeyArg.TestType, // -testtype
                    KeyArg.TimeFrame // -timeframe
                };

                //System.Threading.Thread.Sleep(10000);

                for (int i = 1; i <= rawArgs.Length; i++)
                {
                    bool isArgFinded = false;

                    foreach (var keyArg in keysArg)
                    {
                        if (IsGetArg(keyArg, rawArgs, ref i) && !isArgFinded)
                            isArgFinded = true;
                    }

                    if (isArgFinded)
                        i--; // аргументы командной строки расположены не по порядку
                }

                foreach (var arg in args)
                {
                    switch(arg.Key)
                    {
                        case KeyArg.Security:
                            SecurityTextBox.Text = arg.Value;
                            break;

                        case KeyArg.From:
                            date = GetParamDateTime(arg.Value.Split(','));
                            StartDatePicker.SelectedDate = new DateTime(date[TimeFrame.Year], date[TimeFrame.Month], date[TimeFrame.Day]);
                            break;

                        case KeyArg.To:
                            date = GetParamDateTime(arg.Value.Split(','));
                            EndDatePicker.SelectedDate = new DateTime(date[TimeFrame.Year], date[TimeFrame.Month], date[TimeFrame.Day]);
                            break;

                        case KeyArg.TestType:
                            if (arg.Value == TestType.Candle.ToString())
                                testType = TestType.Candle;
                            break;

                        case KeyArg.TimeFrame:
                            if (testType == TestType.Candle)
                            {
                                //candleTimeFrame = arg.Value; //TODO: 1m 1h 1d 1w 1m 1y
                                candleTimeFrame = TimeSpan.FromMinutes(1);
                            }
                            break;
                    }
                }

                /*if (testType == TestType.Candle.ToString())
                    SetAdditionalTestArgument(TestType.Candle);

                if (testType == TestType.Tick.ToString())
                    SetAdditionalTestArgument(TestType.Tick);

                if (testType == TestType.Depth.ToString())
                    SetAdditionalTestArgument(TestType.Depth);

                if (testType == TestType.Level1.ToString())
                    SetAdditionalTestArgument(TestType.Level1);

                if (testType == TestType.OrderLog.ToString())
                    SetAdditionalTestArgument(TestType.OrderLog);

                if (testType == TestType.CandleAndDepth.ToString())
                    SetAdditionalTestArgument(TestType.CandleAndDepth);

                if (testType == TestType.TickAndDepth.ToString())
                    SetAdditionalTestArgument(TestType.TickAndDepth);
                */

                UseMultibootMenuItem.IsChecked = false;

                SaveBinDataButton_Click(sender, e);
            }
        }

        private Dictionary<TimeFrame, int> GetParamDateTime(string[] rawDate)
        {
            return new Dictionary<TimeFrame, int>
            {
                [TimeFrame.Year] = Convert.ToInt16(rawDate[0]),
                [TimeFrame.Month] = Convert.ToInt16(rawDate[1]),
                [TimeFrame.Day] = Convert.ToInt16(rawDate[2])
            };
        }

        private void SetAdditionalTestArgument(TestType testType)
        {
            switch (testType)
            {
                case TestType.Candle:
                    break;

                case TestType.Tick:
                    break;

                case TestType.Depth:
                    break;

                case TestType.Level1:
                    break;

                case TestType.OrderLog:
                    break;

                case TestType.CandleAndDepth:
                    break;

                case TestType.TickAndDepth:
                    break;

                default:
                    throw new Exception("Unknown test type");
            }
        }

        private string GetArg(string keyArg, string[] rawArgs, ref int i)
        {
            if (i >= rawArgs.Length)
                return null; // arguments are end

            if (rawArgs[i] == "-" + keyArg)
            {
                string valueArg;
                i++;
                valueArg = rawArgs[i];
                i++;
                return valueArg;
            }

            return null;
        }

        private bool IsGetArg(KeyArg keyArg, string[] rawArgs, ref int i)
        {
            if (!args.ContainsKey(keyArg))
            {
                string valueArg = GetArg(keyArg.ToString().ToLower(), rawArgs, ref i);
                if (valueArg != null)
                {
                    args.Add(keyArg, valueArg);
                    return true;
                }
            }

            return false;
        }

        private void WriteMultibootList(string xmlPath, List<Stock> securities)
        {
            var serializer = new XmlSerializer(typeof(List<Stock>));

            if (!File.Exists(xmlPath))
            {
                File.Create(xmlPath).Close();
            }

            using (StreamWriter sw = new StreamWriter(xmlPath, false, Encoding.Unicode))
            {
                serializer.Serialize(sw, securities);
            }
        }

        private bool IsUpdateMultibootList(string xmlPath, string newSecurityId, ref List<Stock> securities)
        {
            if (newSecurityId == "")
                return false;

            securities = ReadMultibootList(xmlPath);

            foreach (Stock security in securities)
            {
                if (security.Id == newSecurityId)
                    return false;
            }

            return true;
        }

        private List<Stock> ReadMultibootList(string multibootXmlPath)
        {
            var binDataLoaderDirectory = GetBinDataLoaderDirectory(Process.GetProcessById(Process.GetCurrentProcess().Id).MainModule.FileName.Split('\\'));

            var serializer = new XmlSerializer(typeof(List<Stock>));

            if (!File.Exists(binDataLoaderDirectory + multibootXmlPath))
            {
                File.Create(binDataLoaderDirectory + multibootXmlPath).Close();

                using (StreamWriter sw = new StreamWriter(binDataLoaderDirectory + multibootXmlPath, false, Encoding.Unicode))
                {
                    List<Stock> securities = new List<Stock>();
                    serializer.Serialize(sw, securities);
                }
            }

            using (StreamReader sr = new StreamReader(binDataLoaderDirectory + multibootXmlPath))
            {
                return (List<Stock>)serializer.Deserialize(sr);
            }
        }

        private string GetBinDataLoaderDirectory(string[] binDataLoaderPath)
        {
            string binDataLoaderDirectory = "";
            for (int i = 0; i < binDataLoaderPath.Length; i++)
            {
                if ((binDataLoaderPath[i] == "BinDataLoader.exe") || (binDataLoaderPath[i] == "BinDataLoader.vshost.exe"))
                    return binDataLoaderDirectory;
                binDataLoaderDirectory += binDataLoaderPath[i] + @"\";
            }
            throw new Exception("BinDataLoader.exe not found");
        }

        private void OpenBinStorageButton_Click(object sender, RoutedEventArgs e)
        {
            Process.Start(BinStoragePathLabel.Content.ToString());
        }

        private void HistorySourceComboBox_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            UseMultibootMenuItem_Checked(sender, e);
        }

        private void SecuritySourceComboBox_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            UseMultibootMenuItem_Checked(sender, e);
        }

        private void SaveBinDataButton_Click(object sender, RoutedEventArgs e)
        {
            var securities = new List<Security>();
            var xmlSecurities = new List<Stock>();
            var multibootListPath = @"MultibootSource\" + HistorySourceComboBox.SelectedItem + @"\" + SecuritySourceComboBox.SelectedItem;

            if (UseMultibootMenuItem.IsChecked) // Multiboot Mode
            {
                xmlSecurities = ReadMultibootList(multibootListPath);
                foreach (Stock xmlSecurity in xmlSecurities)
                {
                    securities.Add(GetSecurity(xmlSecurity));
                }

                var downloadHistoryData = DownloadHistoryCandles((HistorySource)HistorySourceComboBox.SelectedItem, securities, CandleType.TimeFrame, candleTimeFrame, (DateTime)StartDatePicker.SelectedDate, (DateTime)EndDatePicker.SelectedDate, BootMode.Milti);

                for (int i = 0; i < securities.Count; i++)
                {
                    var security = securities[i];
                    // создаём локальное bin хранилище
                    var candleStorage = _storageRegistry.GetCandleStorage(typeof(TimeFrameCandle), security, candleTimeFrame);
                    // сохраняем загруженные исторические данные в локальное bin хранилище
                    candleStorage.Save(downloadHistoryData[i]);
                }

                MessageBox.Show("Все инструменты сохранёны в " + HistorySourceComboBox.SelectedItem + @"\" + SecuritySourceComboBox.SelectedItem, "Сохранение", MessageBoxButton.OK, MessageBoxImage.Information);
            }
            else // Singleboot Mode
            {
                var newSecurity = GetSecurity(SecurityTextBox.Text.Split('@'));
                securities.Add(newSecurity);

                var downloadHistoryData = DownloadHistoryCandles((HistorySource)HistorySourceComboBox.SelectedItem, securities, CandleType.TimeFrame, candleTimeFrame, (DateTime)StartDatePicker.SelectedDate, (DateTime)EndDatePicker.SelectedDate, BootMode.Single);

                // создаём локальное bin хранилище
                var candleStorage = _storageRegistry.GetCandleStorage(typeof(TimeFrameCandle), securities[0], candleTimeFrame);
                // сохраняем загруженные исторические данные в локальное bin хранилище
                candleStorage.Save(downloadHistoryData[0]);

                MessageBox.Show(String.Format("Инструмент {0} c таймфеймом {1} сохранен в Bin Storage", securities[0], candleTimeFrame), "Сообщение", MessageBoxButton.OK, MessageBoxImage.Information);

                if (IsUpdateMultibootList(multibootListPath, SecurityTextBox.Text, ref xmlSecurities) == true)
                {
                    string securityName = null;
                    while (true)
                    {
                        var securityNameWindow = new SecurityNameWindow();
                        if (securityNameWindow.ShowDialog() == true)
                        {
                            securityNameWindow.WarningGroupBox.Content = "This security is not added to " + HistorySourceComboBox.SelectedItem + @"\" + SecuritySourceComboBox.SelectedItem + ". Set the security name.";
                            securityName = securityNameWindow.SecurityNameTextBox.Text;
                            break;
                        }
                    }
                    xmlSecurities.Add(new Stock(securityName, newSecurity.Id, newSecurity.Board.ToString()));
                    WriteMultibootList(multibootListPath, xmlSecurities);

                    MessageBox.Show("Инструмент с id " + SecurityTextBox.Text + @" сохранён в Source\Security.xml", "Сохранение", MessageBoxButton.OK, MessageBoxImage.Information);
                }

                if (e.RoutedEvent.Name == "Loaded")
                    this.Close(); // загрузчик был вызван тестером
            }
        }

        private List<IEnumerable<TimeFrameCandleMessage>> DownloadHistoryCandles(HistorySource historySource, List<Security> securities, CandleType candleType, TimeSpan timeFrame, DateTime from, DateTime to, BootMode bootMode, bool isSaveRequestedData = false)
        {
            switch (historySource)
            {
                case HistorySource.Finam:
                    // Источник для скачивания с Finam
                    var historyFinamSource = new FinamHistorySource(new InMemoryNativeIdStorage(), new InMemoryExchangeInfoProvider());
                    return GetHistoryCandles(historyFinamSource, new FinamSecurityStorage(new Security()), securities, timeFrame, from, to, bootMode, LoadingProgressBar, isSaveRequestedData);
                default:
                    MessageBox.Show(String.Format("History Source {0} is unknown", historySource), "Ошибка", MessageBoxButton.OK, MessageBoxImage.Error);
                    return null;
            }
        }

        /// <summary>
        /// Загрузить свечи с <see cref="historySource"/>
        /// </summary>
        /// <param name="historySource">Источник загрузки</param>
        /// <param name="securityStorage"></param>
        /// <param name="securities">Инструменты</param>
        /// <param name="timeFrame">Временной диапазон свечи</param>
        /// <param name="from">Начать загрузку маркет-данных</param>
        /// <param name="to">Завершить загрузку маркет-данных</param>
        /// <param name="bootMode"></param>
        /// <param name="loadingProgressBar"></param>
        /// <param name="isSaveRequestedData">Сохранить запрошенные данные в файл</param>
        /// <returns></returns>
        private List<IEnumerable<TimeFrameCandleMessage>> GetHistoryCandles(FinamHistorySource historySource, FinamSecurityStorage securityStorage, List<Security> securities, TimeSpan timeFrame, DateTime from, DateTime to, BootMode bootMode, ProgressBar loadingProgressBar, bool isSaveRequestedData)
        {
            var step = 1;
            var criteria = new Security();
            var downloadedDataList = new List<IEnumerable<TimeFrameCandleMessage>>();

            if (isSaveRequestedData)
                historySource.DumpFolder = "HistoryData";

            LoadingProgressBar.Dispatcher.Invoke(() => LoadingProgressBar.Value = LoadingProgressBar.Minimum, DispatcherPriority.Background);
            LoadingLabel.Dispatcher.Invoke(() => LoadingLabel.Content = LoadingProgressBar.Value + "%", DispatcherPriority.Background);

            // Выполняем обновление хранилища инструментов
            // Перед добавлением каждого инструмента в хранилище вызывается функция (делегат) isCanceled, если функция возвращает false, то обновление
            // хранилища продолжается, если true, то прерывается
            // При добавлении нового инструмента в хранилище вызывается функция (делегат) newSecurity. В нашем случае этот делегат имеет пустое тело (ничего не делает)
            switch (bootMode)
            {
                case BootMode.Single:
                    var isCanceled = false;
                    historySource.Refresh(securityStorage, criteria, loadedSecurity => { isCanceled = IsUpdateLoadingProgressBar(loadedSecurity, securities[0], step++); }, () => isCanceled);
                    break;

                case BootMode.Milti:
                    //criteria = new Security { Type = SecurityTypes.Future }; загрузка только фьючерсов
                    historySource.Refresh(securityStorage, criteria, loadedSecurity => { UpdateLoadingProgressBar(loadedSecurity, step++); }, () => false);
                    break;

                default:
                    throw new Exception("Unknown bootMode");
            }

            for (int i = 0; i < securities.Count; i++)
            {
                var security = securities[i];
                try
                {
                    var historyCandles = historySource.GetCandles(security, timeFrame, from, to);
                    downloadedDataList.Add(historyCandles);
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Ошибка загрузки инструмента " + security + " (" + ex.Message + ")");
                }
            }

            LoadingProgressBar.Dispatcher.Invoke(() => LoadingProgressBar.Value = LoadingProgressBar.Maximum, DispatcherPriority.Background);
            LoadingLabel.Dispatcher.Invoke(() => LoadingLabel.Content = LoadingProgressBar.Value + "%", DispatcherPriority.Background);

            return downloadedDataList;
        }

        private Security GetSecurity(Stock stock)
        {
            return new Security { Id = stock.Id, Board = GetExchangeBroad(stock.Board.Split(' ')[0]) };
        }

        private Security GetSecurity(string[] securityText)
        {
            return new Security { Id = SecurityTextBox.Text, Board = GetExchangeBroad(securityText[1]) };
        }

        private ExchangeBoard GetExchangeBroad(string board)
        {
            switch (board)
            {
                case "MICEX":
                case "TQBR":
                    return ExchangeBoard.Micex;
                case "FORTS":
                    return ExchangeBoard.Forts;
                default:
                    throw new Exception("Unknown exchange board");
            }
        }

        private void UpdateLoadingProgressBar(Security loadedSecurity, int step)
        {
            //int allSecurityCountOnFinam = 15000;
            //int stepUpdate = allSecurityCountOnFinam / 100;

            //using (StreamWriter sw = new StreamWriter("sec.txt", true))
            //{
            //    sw.WriteLine(loadedSecurity.Id + " || " + loadedSecurity.Board + " || " + loadedSecurity.Name);
            //}

            if (step % 150 == 0)
            {
                LoadingProgressBar.Dispatcher.Invoke(() => LoadingProgressBar.Value++, DispatcherPriority.Background);
                LoadingLabel.Dispatcher.Invoke(() => LoadingLabel.Content = LoadingProgressBar.Value.ToString() + "%", DispatcherPriority.Background);
            }
        }

        private bool IsUpdateLoadingProgressBar(Security loadedSecurity, Security security, int step)
        {
            if (loadedSecurity == security)
                return true;

            UpdateLoadingProgressBar(loadedSecurity, step);
            return false;
        }

        private void OpenMultibootFileButton_Click(object sender, RoutedEventArgs e)
        {
            var path = System.IO.Path.GetFullPath(@OpenMultibootFileButton.Content.ToString());
            path = path.Replace("Open ", @"MultibootSource\");
            Process.Start(path);
        }

        private void MenuItem_Click(object sender, RoutedEventArgs e)
        {
            using (var dialog = new System.Windows.Forms.FolderBrowserDialog())
            {
                dialog.Description = SetBinStoragePathMenuItem.Header.ToString();
                System.Windows.Forms.DialogResult result = dialog.ShowDialog();
                if (result == System.Windows.Forms.DialogResult.OK)
                {
                    this.GuiAsync(() =>
                    {
                        BinStoragePathLabel.Content = dialog.SelectedPath + @"\";
                    });
                }
            }
        }

        private void UseMultibootMenuItem_Checked(object sender, RoutedEventArgs e)
        {
            try
            {
                OpenMultibootFileButton.Content = @"Open " + HistorySourceComboBox.SelectedItem + @"\" + SecuritySourceComboBox.SelectedItem;
            }
            catch (NullReferenceException) { /* возникает при запуске приложения */ }
        }
    }
}
